import re
from typing import Optional, Tuple

from .countries import COUNTRY_MAP


def is_vies_member(country_code: str) -> bool:
    """
    The abbreviations to use are the ISO codes, except for Greece and the United Kingdom
    """
    if not re.match(r"^[a-zA-Z]{2}$", country_code):
        raise ValueError("Invalid code")

    country_code = country_code.upper()
    if country_code == "GR":
        # http://publications.europa.eu/code/pdf/370000en.htm
        raise ValueError("Use EL for Greece")

    if country_code == "GB":
        # https://ec.europa.eu/taxation_customs/sites/taxation/files/use_of_gb_and_xi_codes_guidance.pdf
        raise ValueError(
            "Use XI for Northern Ireland and XU for United Kingdom (excluding Northern Ireland)"
        )

    return country_code in COUNTRY_MAP


def parse_number(value: str) -> Optional[Tuple[str, str]]:
    value = value.replace(" ", "").upper()
    country_code = value[:2]
    number = value[2:]

    validator = COUNTRY_MAP.get(country_code, None)
    if validator:
        if validator.match(value):
            return country_code, number

    return None



import re


# Patterns: https://ec.europa.eu/taxation_customs/vies/faq.html#item_11
COUNTRY_MAP = {
    # Austria
    "AT": re.compile(r"^ATU\d{8}$"),
    # Belgium
    "BE": re.compile(r"^BE[0,1]\d{9}$"),
    # Bulgaria
    "BG": re.compile(r"^BG\d{9,10}$"),
    # Cyprus
    "CY": re.compile(r"^CY\d{8}[A-Z]$"),
    # Czech Republic
    "CZ": re.compile(r"^CZ\d{8,10}$"),
    # Germany
    "DE": re.compile(r"^DE\d{9}$"),
    # Denmark
    "DK": re.compile(r"^DK\d{8}$"),
    # Estonia
    "EE": re.compile(r"^EE\d{9}$"),
    # Greece
    "EL": re.compile(r"^EL\d{9}$"),
    # Spain - # The first and last characters may be alpha or numeric; but they may not both be numeric.
    "ES": re.compile(r"^ES[A-Z0-9]\d{7}[A-Z0-9]$"),
    # Finland
    "FI": re.compile(r"^FI\d{8}$"),
    # France
    "FR": re.compile(r"^FR[A-Z0-9]{2}\d{9}$"),
    # Croatia
    "HR": re.compile(r"^HR\d{11}$"),
    # Hungary
    "HU": re.compile(r"^HU\d{8}$"),
    # Ireland
    "IE": re.compile(r"^IE\d[A-Z0-9\+\*]\d{5}[A-Z]{1,2}$"),
    # Italy
    "IT": re.compile(r"^IT\d{11}$"),
    # LT
    "LT": re.compile(r"^LT(\d{9}|\d{12})$"),
    # Luxembourg
    "LU": re.compile(r"^LU\d{8}$"),
    # Latvia
    "LV": re.compile(r"^LV\d{11}$"),
    # Malta
    "MT": re.compile(r"^MT\d{8}$"),
    # Netherlands
    "NL": re.compile(r"^NL[A-Z0-9\+\*]{12}$"),
    # Poland
    "PL": re.compile(r"^PL\d{10}$"),
    # Portugal
    "PT": re.compile(r"^PT\d{9}$"),
    # Romania
    "RO": re.compile(r"^RO\d{2,10}$"),
    # Sweden
    "SE": re.compile(r"^SE\d{12}$"),
    # Slovenia
    "SI": re.compile(r"^SI\d{8}$"),
    # Slovakia
    "SK": re.compile(r"^SK\d{10}$"),
    # Northern Ireland
    "XI": re.compile(r"^XI(?:\d{9}(?:\d{3})?|GD\d{3}|HA\d{3})$"),
    # "GB" United Kingdom (left on 31 January 2020)
}



class Error(Exception):
    """pass"""


class ServiceDown(Error):
    """pass"""


class InvalidInput(Error):
    """pass"""


class InvalidRequesterInput(Error):
    """pass"""


class ServiceUnavailable(Error):
    """pass"""


class MSUnavailable(Error):
    """pass"""


class Timeout(Error):
    """pass"""


class VATBlocked(Error):
    """pass"""


class IPBlocked(Error):
    """pass"""


class GlobalMaxConcurrentReq(Error):
    """pass"""


class GlobalMaxConcurrentReqTime(Error):
    """pass"""


class MSMaxConcurrentReq(Error):
    """pass"""


class MsGlobalMaxConcurrentReqTime(Error):
    """pass"""


class OtherError(Error):
    """pass"""


EXCEPTION_MAP = {
    "INVALID_INPUT": InvalidInput,
    "INVALID_REQUESTER_INFO": InvalidRequesterInput,
    "SERVICE_UNAVAILABLE": ServiceUnavailable,
    "MS_UNAVAILABLE": MSUnavailable,
    "TIMEOUT": Timeout,
    "VAT_BLOCKED": VATBlocked,
    "IP_BLOCKED": IPBlocked,
    "GLOBAL_MAX_CONCURRENT_REQ": GlobalMaxConcurrentReq,
    "GLOBAL_MAX_CONCURRENT_REQ_TIME": GlobalMaxConcurrentReqTime,
    "MS_MAX_CONCURRENT_REQ": MSMaxConcurrentReq,
    "MS_MAX_CONCURRENT_REQ_TIME": MsGlobalMaxConcurrentReqTime,
}



from typing import Tuple, Optional, Dict, Callable

import zeep
from zeep.exceptions import Fault, XMLSyntaxError
from zeep.helpers import serialize_object
from .exceptions import EXCEPTION_MAP, OtherError, ServiceDown
from requests.exceptions import ConnectionError
from zeep.cache import SqliteCache
from zeep.transports import Transport


PROD_WSDL = "https://ec.europa.eu/taxation_customs/vies/checkVatService.wsdl"
TEST_WSDL = "https://ec.europa.eu/taxation_customs/vies/checkVatTestService.wsdl"

Response = Tuple[bool, Optional[Dict]]


class Client(object):
    def __init__(
        self,
        wsdl: str = PROD_WSDL,
        cache_path: Optional[str] = None,
        cache_timeout: Optional[int] = None,
    ) -> None:
        try:
            cache = SqliteCache(path=cache_path, timeout=cache_timeout)
            transport = Transport(cache=cache)
            self.zeep = zeep.Client(wsdl, transport=transport)

        except ConnectionError as e:
            # Failed to fetch wsdl
            raise ServiceDown(str(e))

    def _request(self, method: Callable, **arguments) -> Response:
        try:
            response = method(**arguments)
            return response.valid, serialize_object(response, dict)
        except Fault as error:
            try:
                raise EXCEPTION_MAP[error.message]
            except KeyError:
                raise OtherError
        except XMLSyntaxError as e:
            raise ServiceDown(str(e))

    def check(self, country_code: str, vat_number: str) -> Response:
        return self._request(
            self.zeep.service.checkVat, countryCode=country_code, vatNumber=vat_number
        )

    def check_approx(
        self,
        country_code: str,
        vat_number: str,
        trader_name: Optional[str] = None,
        trader_company_type: Optional[str] = None,  # tns1:traderCompanyType @TODO
        trader_street: Optional[str] = None,
        trader_postcode: Optional[str] = None,
        trader_city: Optional[str] = None,
        requester_country_code: Optional[str] = None,
        requester_vat_number: Optional[str] = None,
    ) -> Response:

        return self._request(
            self.zeep.service.checkVatApprox,
            countryCode=country_code,
            vatNumber=vat_number,
            traderName=trader_name,
            traderCompanyType=trader_company_type,
            traderStreet=trader_street,
            traderPostcode=trader_postcode,
            traderCity=trader_city,
            requesterCountryCode=requester_country_code,
            requesterVatNumber=requester_vat_number,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['viesvatcheck']

package_data = \
{'': ['*']}

install_requires = \
['zeep>=4.2.1,<5.0.0']

setup_kwargs = {
    'name': 'viesvatcheck',
    'version': '0.4.0',
    'description': '',
    'long_description': 'vies-vat-check\n==============================================\n\nPython client for https://ec.europa.eu/taxation_customs/vies/#/technical-information',
    'author': 'Mārtiņš Šulcs',
    'author_email': 'shulcsm@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
